/*
 enchant.jsでライフゲーム
 2012/01/29 @liarnose

 enchant.jsを出力描写とかタイマーとかに使ったライフゲームです
　単にドットを塗るだけだと味気ないので各セルがスプライトになっててわちゃわちゃ動きます

　code9leap先生にカッコがおかしいと言われるのでどっかで構文を間違えている可能性あり
　単に妙なインデントとの相性が悪いだけという可能性もあり（LifeGame.updateがあやしい）
 せっかくあるのにtl.enchant.js丸無視だけどそれでいいのか　いいか　まあ
*/

enchant();

//グローバルな定数
var GAME_SIZE = 320; //描写枠の大きさ
//さあ始まるザマスよ
//いくでガンス！
//フンガー
window.onload = function() {
    game = new Game(GAME_SIZE, GAME_SIZE);
    game.preload('chara1.png');
    game.onload = function() {
        game.pushScene(new LifeGame());
    };
    game.start();
};

//関数スコープでLifeGameのコンストラクタ以外を隠蔽
var LifeGame = (function namespace() {

    //モジュール内定数
    var CELL_SIZE = 16; //セルの大きさ
    var N = GAME_SIZE / CELL_SIZE; //一列あたりのセルの数
    var TIMER = 30; //一世代の時間
    //ライフゲームを走らせるシーン
    //今回は1個あれば十分だから別にクラスにしなくてもいい気がするけど一応
    var LifeGame = Class.create(Scene, {
        initialize: function() { //準備
            Scene.call(this);
            this.timer = 0;
            this._initCells();
            this.sowRPentomino();
        },
        _initCells: function() { //N*Nに並んだセルを生成
            //各セルを行列っぽく見に行きたいので各列をグループにする
            var g;

            for (var i = 0; i < N; i++) {
                g = new Group();
                g.x = (i - 1 / 2) * CELL_SIZE;
                for (var j = 0; j < N; j++) {
                    g.addChild(new Cell(0, (j - 1 / 2) * CELL_SIZE, CELL_SIZE));
                }
                this.addChild(g);
            }
        },
        getCell: function(px, py) { //行と列を指定してライフゲーム内のセルを得る
            if (0 > px || px >= N || 0 > py || py >= N) { //範囲外ならnull
                return null;
            } else {
                return this.childNodes[px].childNodes[py];
            }
        },
        aliveNeighbor: function(px, py) { //行と列を指定してその周りの生きてるご近所さんセルを数える
            //多分重さの元凶なのでかぞえに行かずに逐一変化を記録するとかの方がよさそう
            var ret = 0;
            var A = [{
                x: -1,
                y: -1
            }, {
                x: 0,
                y: -1
            }, {
                x: 1,
                y: -1
            }, {
                x: -1,
                y: 0
            }, {
                x: 1,
                y: -0
            }, {
                x: -1,
                y: 1
            }, {
                x: 0,
                y: 1
            }, {
                x: 1,
                y: 1
            }]; //周り8マス
            for (var n in A) {
                var c = this.getCell(px + A[n].x, py + A[n].y);
                if (c && c.state.isAlive()) {
                    ret++;
                }
            }

            return ret;
        },
        sowRPentomino: function() { //ライフゲームの中央にRペントミノ型の生きてるセルを配置
            var A = [{
                x: -1,
                y: -1
            }, {
                x: 0,
                y: -1
            }, {
                x: 0,
                y: 0
            }, {
                x: 1,
                y: 0
            }, {
                x: 0,
                y: 1
            }]; //Rペントミノの形
            for (var n in A) {
                var c = this.getCell(N / 2 + A[n].x - 1, N / 2 + A[n].y - 1);
                //長生きするように位置を微調整してます
                c.state = new S_Alive();
                c.state.initCell(c);
            }
        },
        onenterframe: function() { //毎フレーム実行される
            this.timer--;
            if (this.timer <= 0) {
                this.update();
                this.timer = TIMER;
            }
        },
        update: function() { //世代更新作業 イテレータ書いたほうが  
            for (var x = 0; x < N; x++) {
                for (var y = 0; y < N; y++) {
                    var c = this.getCell(x, y);
                    var n = this.aliveNeighbor(x, y);
                    c.poststate = c.state.nextState(n);
                }
            }
            for (var x = 0; x < N; x++) {
                for (var y = 0; y < N; y++) {
                    var c = this.getCell(x, y);
                    if (c.poststate && c.poststate != c.state) {
                        c.state = c.poststate;
                        c.state.initCell(c);
                    }
                }
            }
        }
    });

    //セルの持つ状態たち　ダックタイピングなStateパターン
    var S_Dead = Class.create({ //セルの状態　死亡
        nextState: function(n) { //生きてるご近所数ごとの次の状態
            if (n == 3) {
                return new S_Alive();
            } else {
                return this;
            }
        },
        initCell: function(cell) { //状態が移行したときにやる処理
            //人手で呼ばせてるけどセルに状態変更があったことを検知させて自動で呼ばれるようにした方がいいのかも
            cell.opacity = 0;
            cell.frame = 3;
            cell.onenterframe = function() { //特に毎フレームやる処理はないので空の関数でリセット
            };
        },
        isAlive: function() { //生きてますかー？
            return false;
        }
    });
    var S_Dying = Class.create({ //セルの状態　今死んだ
        nextState: function(n) { //生きてるご近所数ごとの次の状態
            if (n == 3) {
                return new S_Alive();
            } else {
                return new S_Dead();
            }
        },
        initCell: function(cell) { //状態が移行したときにやる処理
            cell.opacity = 0.5;
            cell.frame = 3;
            cell.onenterframe = function() { //フェードアウトする処理
                this.opacity *= 0.8;
            };
        },
        isAlive: function() {
            return false;
        }
    });
    var S_Alive = Class.create({ //セルの状態　生存
        initialize: function() {
            this.timersx = 0;
            this.timerf = 0;
        },
        nextState: function(n) { //生きてるご近所数ごとの次の状態
            if (n == 2 || n == 3) {
                return this;
            } else {
                return new S_Dying();
            }
        },
        initCell: function(cell) {
            cell.opacity = 1;
            cell.frame = 0;
            cell.onenterframe = function() { //わちゃわちゃ動く処理
                this.state.timersx--;
                if (this.state.timersx <= 0) {
                    this.scaleX *= -1;
                    this.state.timersx = Math.floor(Math.random() * 20);
                }

                this.state.timerf--;
                if (this.state.timerf <= 0) {
                    this.frame = (this.frame + 1) % 3;
                    this.state.timerf = Math.floor(Math.random() * 3 + 2);
                }
            };
        },
        isAlive: function() {
            return true;
        }
    });

    //ライフゲームの個々の細胞
    //ある程度アニメーションさせるつもりだったのでSpriteだけど
    //単にライフゲームをやるだけならセルを全部まとめてMapとした方がお手軽そう
    Cell = Class.create(Sprite, {
        initialize: function(x, y, size) { //初期化する
            Sprite.call(this, 32, 32);
            this.image = game.assets['chara1.png'];
            this.x = x;
            this.y = y;
            this.scaleX = size / 32;
            this.scaleY = size / 32;
            this.frame = 0;
            this.poststate = new S_Dying(); //次の状態　更新作業用
            this.state = new S_Dead(); //初期状態ではセルは死んでる
            this.state.initCell(this);
        }
    });

    return LifeGame;
}());